import unittest

from ratisbona_utils.strings import (
    snake_to_words, words_to_camel, camel_to_words, pascal_to_words, shorten, indent
)
from ratisbona_utils.strings._alignment import si_format_number


class TestStrings(unittest.TestCase):

    def test_snake_to_words_must_yield_words(self):
        words = snake_to_words('this_is_a_test')
        self.assertEqual(words, ['this', 'is', 'a', 'test'])

    def test_words_to_camel_must_yield_camel_casing(self):
        result = words_to_camel(['this', 'is', 'a', 'test'])
        self.assertEqual('thisIsATest', result)

    def test_camel_to_words_must_yield_words_in_lowercase(self):
        result = camel_to_words('entityFactoryProviderInterface')
        self.assertEqual(['entity','factory', 'provider', 'interface'], result)

    def test_camel_to_words_must_yield_word_if_only_one_word_given(self):
        result = camel_to_words('entity')
        self.assertEqual(['entity'], result)

    def test_camel_to_words_must_complain_if_not_camel_given(self):
        with self.assertRaises(ValueError) as ve:
            camel_to_words('EntityProvider')
        self.assertTrue('camel' in str(ve.exception))
        self.assertTrue('not' in str(ve.exception))

    def test_pascal_to_words_must_yield_words_in_lowercase(self):
        result = pascal_to_words('EntityFactoryProviderInterface')
        self.assertEqual(['entity','factory', 'provider', 'interface'], result)

    def test_pascal_to_words_must_yield_word_if_only_one_word_given(self):
        result = pascal_to_words('Entity')
        self.assertEqual(['entity'], result)

    def test_camel_to_words_must_complain_if_not_pascal_given(self):
        with self.assertRaises(ValueError) as ve:
            camel_to_words('EntityProvider')
        self.assertTrue('camel' in str(ve.exception))
        self.assertTrue('not' in str(ve.exception))

    def test_pascal_to_words_must_complain_if_not_pascal_given(self):
        with self.assertRaises(ValueError) as ve:
            pascal_to_words('entityProvider')
        self.assertTrue('pascal' in str(ve.exception))
        self.assertTrue('not' in str(ve.exception))

    def test_shorten_must_return_text_if_shorter_than_max_length(self):
        result = shorten('this is a test', 20)
        self.assertEqual('this is a test', result)

    def test_shorten_must_raise_if_max_length_less_than_14(self):
        with self.assertRaises(ValueError) as ve:
            shorten('this is a test', 10)
        self.assertTrue('14' in str(ve.exception))

    def test_shorten_must_shorten(self):
        test='this is a very long test with many chars. Hello World, how are you?'
        for i in range(14, 20):
            result = shorten(test, i)
            print(result)
            self.assertTrue(len(result) == i)

    def test_elipsis_must_contain_correct_string_length(self):
        for strlen in range(20, 1000, 10):
            string = ("ABCDEFGHIJKLMNOPQRSTUVWXYZ" * (strlen // 26 + 1))[:strlen]

            for i in range(14, 19 if strlen < 101 else 100):
                result = shorten(string, i)
                print(result)
                before, _, after = result.split("...")
                self.assertTrue(string.startswith(before))
                self.assertTrue(string.endswith(after))
                self.assertTrue(len(result) == i)
                self.assertTrue(f"[{len(string)}]" in result)

    def test_elipsis_must_contain_999plus_for_very_long_strings(self):
        string = "A" * 1000
        result = shorten(string, 14)
        print(result)
        self.assertTrue(len(result) == 14)
        self.assertTrue("[>999]" in result)

    def test_indent_must_indent(self):
        # given
        text = 'This is a test\nwith two lines'
        # when
        result = indent(text, 4)
        # then
        self.assertEqual('    This is a test\n    with two lines', result)



    def test_si_format_number_decimal(self):
        self.assertEqual(si_format_number(999, binary_units=False), '999')
        self.assertEqual(si_format_number(1000, binary_units=False), '1 k')
        self.assertEqual(si_format_number(1500, binary_units=False), '1 k')
        self.assertEqual(si_format_number(1000000, binary_units=False), '1 M')
        self.assertEqual(si_format_number(1000000000, binary_units=False), '1 G')
        self.assertEqual(si_format_number(1000000000000, binary_units=False), '1 T')

    def test_si_format_number_binary(self):
        self.assertEqual(si_format_number(1023, binary_units=True), '1023')
        self.assertEqual(si_format_number(1024, binary_units=True), '1 ki')
        self.assertEqual(si_format_number(1536, binary_units=True), '1 ki')
        self.assertEqual(si_format_number(1048576, binary_units=True), '1 Mi')
        self.assertEqual(si_format_number(1073741824, binary_units=True), '1 Gi')
        self.assertEqual(si_format_number(1099511627776, binary_units=True), '1 Ti')


if __name__ == '__main__':
    unittest.main()



import re
from enum import Enum, auto
from itertools import zip_longest
import textwrap


class Alignment(Enum):
    """
    Base class for enums.
    """

    LEFT = auto()
    RIGHT = auto()
    CENTER = auto()
    BLOCK = auto()


ALIGNMENT_FUNCTIONS = {
    Alignment.LEFT: str.ljust,
    Alignment.RIGHT: str.rjust,
    Alignment.CENTER: str.center,
    Alignment.BLOCK: lambda text, width: text,
}


def align(text: str, width=80, alignment=Alignment.LEFT) -> str:
    """
    Align text in a given width.

    Args:
        text: The text to align.
        width: The width to align the text in.
        alignment: The alignment to use.

    Returns:
        str, the aligned text.
    """
    return ALIGNMENT_FUNCTIONS[alignment](text, width)


def longest_line(multiline_text: str) -> int:
    return max(len(line) for line in multiline_text.splitlines())


def zip_longest_textcolumn(*args: str, separator: str = " ") -> str:
    """
    Given several multiline texts (as lists of lines), this function will zip them together as columns, which each
    text you gave forming one column. The columns will be aligned to the left.

    Args:
        *args: The multiline texts.

        separator: The separator to use between the columns.

    Returns:
        The multiline text arranged as multiple columns.
    """

    colwidths = [longest_line(arg) for arg in args]

    split_args = [arg.splitlines() for arg in args]
    result = ""
    for idx, lines in enumerate(zip_longest(*split_args, fillvalue="")):
        result += (
            separator.join(line.ljust(width) for line, width in zip(lines, colwidths))
            + "\n"
        )
    return result


def shorten(
    text: str,
    max_length: int
) -> str:
    """
    Shorten a text to a maximum length, using an elipsis provider to provide the elipsis text.

    Args:
        text: The text to shorten.
        max_length: The maximum length of the text.
        elipsis_provider: A function that takes the length of the text and returns the elipsis text.
        elipsis_length: The length of the elipsis text.
    """
    TOO_LONG_REPLACEMENT="...[>999]..."
    LEAST_LENGTH = len(TOO_LONG_REPLACEMENT) + 2

    if max_length < LEAST_LENGTH:
        raise ValueError(f"max_length must be at least {LEAST_LENGTH}")

    length = len(text)
    if length <= max_length:
        return text

    elipsis = f"...[{length}]..." if length < 1000 else TOO_LONG_REPLACEMENT
    elipsis_length = len(elipsis)

    length_before_elipsis = (max_length - elipsis_length) // 2
    length_after_elipsis = max_length - elipsis_length - length_before_elipsis
    return text[:length_before_elipsis] + elipsis + text[-length_after_elipsis:]


def indent(text: str, indent: int) -> str:
    """
    Indent a text by a given number of spaces.

    Args:
        text: The text to indent.
        indent: The number of spaces to indent the text by.

    Returns:
        The indented text.
    """
    return "".join([
        f"{' ' * indent}{line}"
        for line in text.splitlines(keepends=True)
    ])

def rewrap_text(text: str, width: int) -> str:
    """Rewraps a string to a given column width, preserving existing double newlines.

    Args:
        text (str): The input text to rewrap.
        width (int): The maximum column width.

    Returns:
        str: The rewrapped text.
    """
    longest_line = max(len(line) for line in text.splitlines())
    if longest_line <= width:
        return text
    # Replace all single newlines in text by space. Keep double newlines.
    text = re.sub(r'(?<!\n)\n(?!(\n|\s))', ' ', text)
    return wrap_text(text, width)


def wrap_text(text, width):
    """Wraps a string to a given column width, preserving existing newlines.

    Args:
        text (str): The input text to wrap.
        width (int): The maximum column width.

    Returns:
        str: The wrapped text.
    """
    wrapped_lines = []

    for line in text.splitlines():  # Preserve existing newlines
        wrapped_lines.extend(
            textwrap.fill(line.strip(), width, break_long_words=True, break_on_hyphens=True).split("\n")
        )

    return "\n".join(wrapped_lines)

def si_format_number(the_number: int, binary_units=True) -> str:
    """
    Formats a number in SI units.

    Args:
        the_number (int): The number to format.
        binary_units (bool): Whether to use binary units (1024) or decimal units (1000).

    Returns:
        str: The formatted number, like '12M' or '12Mi'.
    """
    base = 1024 if binary_units else 1000
    postfix = ['',' k',' M',' G',' T']
    times = 0
    while the_number >= base and times < len(postfix):
        the_number = the_number // base
        times += 1
    return f'{the_number}{postfix[times]}' + ('i' if binary_units and times > 0 else '')




import re
from functools import partial
from itertools import pairwise
from typing import Sequence


def snake_to_words(word: str) -> list[str]:
    return list(map(lambda _: _.lower(), word.split('_')))


def words_to_snake(words: Sequence[str]) -> str:
    return '_'.join(map(lambda w: w.lower(), words))


def _camel_or_pascal_to_words(word: str, is_pascal: bool) -> list[str]:
    # Bracketing the matcher in re.split causes the delimiter to become
    # a match group of it's own:
    # a=re.split('(-)', '1-2-3') #=> ['1', '-', '2', '-', '3']
    head, *delimited = re.split(r'([A-Z]+)', word)
    # For Pascal casing (starting with a Uppercase letter, the head must always be an empty group.
    # For Camel casing the head must be filled. Any other combination is an error.
    if is_pascal ^ (head == ''):
        raise ValueError(f'{word}: Thats not {"pascal" if is_pascal else "camel"} casing.')
    words = list(pairwise(delimited))[::2]
    words = map(''.join, words)
    if not is_pascal:
        words = [head, *words]
    words = map(str.lower, words)
    return list(words)


camel_to_words = partial(_camel_or_pascal_to_words, is_pascal=False)

pascal_to_words = partial(_camel_or_pascal_to_words, is_pascal=True)


def words_to_camel(words: Sequence[str]) -> str:
    head, *tail = words
    return ''.join([head, *words_to_pascal(tail)])


def words_to_pascal(words: Sequence[str]) -> str:
    return ''.join(map(lambda w: w.capitalize(), words))


import inspect
import shutil
from datetime import datetime
from pathlib import Path

from matplotlib import colormaps

import importlib.resources as pkg_resources
import click

import ratisbona_utils
import ratisbona_utils.colors.good_palettes as good_palettes
from ratisbona_utils.boxdrawing import blue_dosbox
from ratisbona_utils.colors import (
    parse_gimp_palette,
    to_gimp_palette,
    palette_generators,
    Palette,
    rgb_to_hex, to_gimp_gradient, rgb_to_normalized,
)
from ratisbona_utils.colors.palette_generators import matplotlib_colormap_to_palette
from ratisbona_utils.colors.serde_json_palette import to_json, from_json
from ratisbona_utils.io import errprint
from ratisbona_utils.terminals.vt100 import color_block


def do_display_palette(palette) -> None:
    for color, name, colortype in zip(
        palette.colors, palette.color_names, palette.color_types
    ):
        #displaycolor = rgb_to_hex(color) if colortype == "rgb" else str(color)
        block = color_block(color)
        print(f"{block}", end="")


def display_named_colors(palette: Palette) -> None:
    for color, name, colortype in zip(
        palette.colors, palette.color_names, palette.color_types
    ):
        displaycolor = rgb_to_hex(color) if colortype == "rgb" else str(color)
        block = color_block(color)
        print(f"{block} {name} {displaycolor}")


@click.group()
def palette_cli():
    errprint(blue_dosbox("Ratisbona Palette CLI"))

@palette_cli.command()
@click.argument(
    "palette_json",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
def incorporate(palette_json: Path):
    """
    Incorporate a palette JSON file into the good_palettes module.
    """
    with pkg_resources.files(ratisbona_utils.colors.resources) as resource_path:
        shutil.copy(palette_json, resource_path / palette_json.name)


@click.command("parse-gimp")
@click.argument(
    "gimp_palette_file",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def gimp_palette_parser(gimp_palette_file: Path, output: Path | None):
    """
    Parse a GIMP palette file and print the palette object.
    """
    palette_str = gimp_palette_file.read_text()
    gimp_palette = parse_gimp_palette(palette_str)
    json_text = to_json(gimp_palette)
    if output:
        output.write_text(json_text)
    else:
        print(json_text)


@palette_cli.command("write-gimp")
@click.argument(
    "native_palette",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def gimp_palette_writer(native_palette: Path, output: Path | None):
    """
    Write a GIMP palette file from a palette object.
    """
    palette_json = native_palette.read_text()
    palette = from_json(palette_json)
    gimp_palette = to_gimp_palette(palette)
    if output:
        output.write_text(gimp_palette)
    else:
        print(gimp_palette)


def identifier_to_palettename(identifier: str) -> str:
    return identifier.replace("_PALETTE", "").replace("_", "-").lower()


def _palettes_dict():
    identifiers = dir(good_palettes)
    palettes = list(filter(lambda identifier: identifier.endswith("_PALETTE"), identifiers))
    palette_names = (map(identifier_to_palettename, palettes))
    palette_instances = (map(lambda identifier: getattr(good_palettes, identifier), palettes))
    return dict(zip(palette_names, palette_instances))

def _resource_palettes():
    palettes = {}
    with pkg_resources.path(ratisbona_utils.colors.resources, '') as resource_path:
        for file in resource_path.iterdir():
            if file.suffix == '.json':
                palette = from_json(file.read_text())
                palettes[palette.name.lower()] = palette
    return palettes


PALETTES = dict(**_palettes_dict(), **_resource_palettes())

@palette_cli.command("writeout-builtin")
@click.argument("palette_name", type=click.Choice(PALETTES))
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def writeout_builtin_palette(palette_name: str, output: Path | None):
    """
    Write a builtin palette to a file.
    """
    palette = PALETTES[palette_name]
    json_str = to_json(palette)
    do_display_palette(palette)
    if output:
        output.write_text(json_str)
    else:
        print(json_str)

GENERATORS = ["cielch_based_rgb_palette", "hsv_based_rgb_palette" ]


@palette_cli.command("writeout-generated")
@click.argument("generator", type=click.Choice(GENERATORS + [map for map in colormaps]))
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
@click.option(
    "--hue-steps",
    "-n",
    type=int,
    required=False,
    default=16,
)
@click.option(
    "--lightness",
    "-l",
    type=float,
    required=False,
    default=50,
)
@click.option(
    "--chroma",
    "-c",
    type=float,
    required=False,
    default=50,
)
@click.option(
    "--saturation",
    "-s",
    type=float,
    required=False,
    default=1,
)
@click.option(
    "--value",
    "-v",
    type=float,
    required=False,
    default=1,
)
def writeout_generated_palette(
    generator, output: Path | None, hue_steps, lightness, chroma, saturation, value
):
    """
    Write a generated palette to a file.

    Args:
        generator:
        output:
        hue_steps:
        lightness:
        chroma:
        saturation:
        value:

    Returns:
    """
    provided_args = {
        "hue_steps": hue_steps,
        "lightness": lightness,
        "chroma": chroma,
        "saturation": saturation,
        "value": value,
    }
    if generator in GENERATORS:
        palette = getattr(palette_generators, generator)
    elif generator in colormaps:
        def palette(hue_steps: int):
            return matplotlib_colormap_to_palette(hue_steps, generator)
    else:
        raise ValueError(f"Unknown generator {generator}")


    signature = inspect.signature(palette)
    valid_args = {
        key: value
        for key, value in provided_args.items()
        if key in signature.parameters
    }
    bound_args = signature.bind_partial(**valid_args)
    rgb_colors = palette(**bound_args.arguments)
    palette = Palette(
        name=f"{generator}_palette",
        description=f"Generated with {generator}",
        author="palette_cli",
        creation_date=datetime.now(),
        color_names=[""] * len(rgb_colors),
        color_types=["rgb"] * len(rgb_colors),
        colors=rgb_colors,
    )
    json_str = to_json(palette)
    do_display_palette(palette)
    if output:
        output.write_text(json_str)
    else:
        print(json_str)


@palette_cli.command("display")
@click.argument(
    "palette_file",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
def display_palette(palette_file: Path):
    """
    Display a palette file.
    """
    palette = from_json(palette_file.read_text())
    display_named_colors(palette)
    do_display_palette(palette)


@palette_cli.command("convert-to-gimp-gradient")
@click.argument(
    "palette_file", type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True)
)
@click.option(
    "--output-ggr-file",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def convert_to_gimp_gradient(palette_file: Path, output_ggr_file: Path):
    """
    Convert a palette file to a GIMP gradient file.
    """
    palette = from_json(palette_file.read_text())
    title = palette.name
    normalized_colors = [rgb_to_normalized(color) for color in palette.colors]
    gimp_gradient = to_gimp_gradient(title, normalized_colors)
    if output_ggr_file:
        output_ggr_file.write_text(gimp_gradient)
    else:
        print(gimp_gradient)


@palette_cli.command("parse-gimp")
@click.argument(
    "gimp_palette_file",
    type=click.Path(exists=True, path_type=Path, dir_okay=False, file_okay=True),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, path_type=Path, dir_okay=False, file_okay=True),
    required=False,
    default=None,
)
def parse_gimp_palette_command(gimp_palette_file: Path, output: Path | None):
    """
    Parse a GIMP palette file and print the palette object.
    """
    palette_str = gimp_palette_file.read_text()
    gimp_palette = parse_gimp_palette(palette_str)
    display_named_colors(gimp_palette)
    do_display_palette(gimp_palette)
    json_text = to_json(gimp_palette)
    if output:
        output.write_text(json_text)
    else:
        print(json_text)


from ratisbona_utils.colors import Palette, RGBColor


def to_sourcecode(palette: Palette[RGBColor]) -> str:

    text = ""
    for color, name in zip(palette.colors, palette.color_names):
        python_name = name.replace("-", "_")
        r, g, b = color
        text += f"{python_name}=({hex(r)}, {hex(g)}, {hex(b)})\n"

    capitalized_palette_name = palette.name.replace(" ", "_").upper()

    text += f"{capitalized_palette_name}=[\n"
    for name in palette.color_names:
        python_name = name.replace("-", "_")
        text += f"    {python_name},\n"
    text += "]"

    print(text)


import unittest
from ratisbona_utils.arrays._slices import slice_2d_recognition, slice_parsing

class SlicesTestCase(unittest.TestCase):

    def test_recognize_slices_must_allow_single_slice(self):
        rowslice, colslice = slice_2d_recognition(slice(2, 5))
        self.assertEqual(rowslice, slice(2, 5))
        self.assertIsNone(colslice)

    def test_recognize_slices_must_allow_tuple_of_slices(self):
        rowslice, colslice = slice_2d_recognition((slice(1, 4), slice(2, 6)))
        self.assertEqual(rowslice, slice(1, 4))
        self.assertEqual(colslice, slice(2, 6))

    def test_recognize_slices_must_raise_value_error_on_invalid_tuple_length(self):
        with self.assertRaises(ValueError) as cm:
            _ = slice_2d_recognition((slice(0, 5), slice(0, 5), slice(0, 5)))
        print(cm.exception)

    def test_recognize_slices_must_allow_single_integer(self):
        rowslice, colslice = slice_2d_recognition(3)
        self.assertEqual(rowslice, 3)
        self.assertIsNone(colslice)

    def test_recognize_slices_must_allow_tuple_of_integer_and_slice(self):
        rowslice, colslice = slice_2d_recognition((4, slice(1, 5)))
        self.assertEqual(rowslice, 4)
        self.assertEqual(colslice, slice(1, 5))

    def test_recognize_slices_must_allow_tuple_of_slice_and_integer(self):
        rowslice, colslice = slice_2d_recognition((slice(0, 3), 2))
        self.assertEqual(rowslice, slice(0, 3))
        self.assertEqual(colslice, 2)

    def test_regonize_slices_must_allow_tuple_of_integers(self):
        rowslice, colslice = slice_2d_recognition((1, 3))
        self.assertEqual(rowslice, 1)
        self.assertEqual(colslice, 3)

    def test_parse_slices_must_handle_slice_correctly(self):
        start, end, step = slice_parsing(slice(2, 8, 3), 10)
        self.assertEqual((start, end, step), (2, 8, 3))

    def test_parse_slices_must_handle_integer_correctly(self):
        start, end, step = slice_parsing(5, 10)
        self.assertEqual((start, end, step), (5, 6, 1))

    def test_parse_slices_must_handle_slice_with_none_values(self):
        start, end, step = slice_parsing(slice(None, None, None), 10)
        self.assertEqual((start, end, step), (0, 10, 1))

    def test_parse_slices_must_raise_value_error_on_invalid_type(self):
        with self.assertRaises(ValueError) as cm:
            _ = slice_parsing("invalid", 10)
        print(cm.exception)

    def test_parse_slices_must_handle_none_as_full_range(self):
        start, end, step = slice_parsing(None, 10)
        self.assertEqual((start, end, step), (0, 10, 1))




if __name__ == '__main__':
    unittest.main()





def make_text_to_latex_paragraphs(text: str) -> str:
    """
    Converts plain text into LaTeX paragraphs by splitting at double newlines.
    Can also be used to remove excessive newlines.

    Args:
        text (str): The input plain text.
    Returns:
        str: The text formatted with LaTeX paragraph breaks.
    """
    paragraphs = text.split("\n")
    latex_paragraphs = [p.strip() for p in paragraphs if p.strip()]
    return "\n\n".join(latex_paragraphs) + "\n"



from pathlib import Path
from unittest import TestCase

from ratisbona_utils.pipes import cat, collect, fcat


class TestPipes(TestCase):

    def test_cat(self):
        result = cat(range(0, 10)) | (lambda x: x * 2) | (lambda x: x + 1) | collect(list)
        print(result)
        self.assertEqual(result, [1, 3, 5, 7, 9, 11, 13, 15, 17, 19])

    def test_fcat(self):
        result = fcat(Path(__file__)) | (lambda x: f'-->{x}<--') | collect(list)
        print(result)

        with open(Path(__file__)) as f:
            expected_content = ['-->' + x + '<--' for x in f.read().splitlines()]

        self.assertEqual(result, expected_content)



from dataclasses import dataclass, field
from typing import TypeVar, Generic, Generator, Iterable, Iterator

from ._model import LineToken

T = TypeVar('T')


@dataclass
class TokenStream(Generic[T]):
    """
    A stream of tokens that can be read and put back for backtracking.

    Args:
        tokenizer: A provider of tokens.
    """
    tokenizer: Iterator[T]
    buffer: list[T] = field(default_factory=list[T])

    def next_token(self):
        """
        Retrieves the next token from the stream.

        Returns:
            The next token, or None if the stream is exhausted.
        """
        if self.buffer:
            return self.buffer.pop()  # Retrieve the most recently put-back token
        return next(self.tokenizer)

    def put_back(self, token):
        """
        Puts a token back into the stream for backtracking.

        Args:
            token: The token to put back.
        """
        self.buffer.append(token)


def line_tokenizer(stream: Iterable[str]) -> Generator[LineToken, None, None]:
    """
    A generator function that yields LineTokens from a given stream.

    Args:
        stream: An iterable stream of lines (e.g., a file object or any iterable of strings).

    Yields:
        LineToken: A token representing a line of text with its line number.
    """
    for line_number, line in enumerate(stream, start=1):
        yield LineToken(text=line.rstrip('\n'), line_number=line_number)


from datetime import date, datetime, time
from unittest import TestCase

from ratisbona_utils.datetime import lenient_tokenizer, leniently_parse, ensure_datetime
from ratisbona_utils.functional import Provider


class TestDateParser(TestCase):

    def do_parse(self, text, now_provider: Provider[datetime] = datetime.now):
        tokens = list(lenient_tokenizer(text))
        for token in tokens:
            print(token)
        result = leniently_parse(tokens, base_date_for_implicit_dates=now_provider)
        print(result)
        return result

    def test_date_parser_must_parse_valid_isodate(self):
        self.assertEqual(date(1979, 7, 18), self.do_parse("1979-07-18")[0])

    def test_date_parser_must_parse_valid_german_date(self):
        self.assertEqual(date(1979, 7, 18), self.do_parse("18.07.1979")[0])

    def test_date_parser_must_parse_valid_american_date(self):
        self.assertEqual(date(1979, 7, 18), self.do_parse("07/18/1979")[0])

    def test_date_parser_must_leniently_parse_valid_isodate_month(self):
        self.assertEqual(date(1979, 7, 1), self.do_parse("1979-7-01")[0])

    def test_date_parser_must_leniently_parse_valid_germandate_month(self):
        self.assertEqual(date(1979, 7, 18), self.do_parse("18.7.1979")[0])

    def test_date_parser_must_leniently_parse_valid_american_date_month(self):
        self.assertEqual(date(1979, 7, 18), self.do_parse("7/18/1979")[0])

    def test_date_parser_must_leniently_parse_valid_isodate_day(self):
        self.assertEqual(date(1979, 7, 1), self.do_parse("1979-07-1")[0])

    def test_date_parser_must_leniently_parse_valid_germandate_day(self):
        self.assertEqual(date(1979, 7, 1), self.do_parse("1.07.1979")[0])

    def test_date_parser_must_leniently_parse_valid_american_date_day(self):
        self.assertEqual(date(1979, 7, 1), self.do_parse("07/1/1979")[0])

    def test_date_parser_must_not_leniently_parse_isoday_year(self):
        self.assertFalse(self.do_parse("79-07-18")[0])

    def test_date_parser_must_leniently_parse_germandate_year_taking_century_from_now_provider(
        self,
    ):
        self.assertEqual(
            date(2025, 7, 18),
            self.do_parse(
                "18.07.25", now_provider=lambda: datetime(2025, 1, 1, 0, 0)
            )[0],
        )
        self.assertEqual(
            date(1925, 7, 18),
            self.do_parse(
                "18.07.25", now_provider=lambda: datetime(1926, 1, 1, 0, 0)
            )[0],
        )

    def test_date_parser_must_leniently_parse_germandate_year_taking_century_from_now_provider_using_previous_century_if_in_future(
        self,
    ):
        self.assertEqual(
            date(1925, 7, 18),
            self.do_parse("18.07.25", now_provider=lambda: datetime(2024, 1, 1, 0, 0))[0],
        )
        self.assertEqual(
            date(1825, 7, 18),
            self.do_parse("18.07.25", now_provider=lambda: datetime(1924, 1, 1, 0, 0))[0],
        )

    def test_date_parser_must_leniently_parse_german_date_without_year_taking_year_from_now_provider(self):
        self.assertEqual(
            date(2025, 7, 18),
            self.do_parse("18.07.", now_provider=lambda: datetime(2025, 1, 1, 0, 0))[0],
        )
        self.assertEqual(
            date(1925, 7, 18),
            self.do_parse("18.07.", now_provider=lambda: datetime(1925, 1, 1, 0, 0))[0],
        )

    def test_date_parser_must_leniently_parse_american_date_year(self):
        self.assertEqual(
            date(2025, 7, 18),
            self.do_parse(
                "07/18/25", now_provider=lambda: datetime(2025, 1, 1, 0, 0)
            )[0],
        )
        self.assertEqual(
            date(1925, 7, 18),
            self.do_parse(
                "07/18/25", now_provider=lambda: datetime(1926, 1, 1, 0, 0)
            )[0],
        )


    def test_date_parser_must_not_accept_illegal_day_isoformat(self):
        self.assertFalse(self.do_parse("1979-07-32")[0])
        self.assertFalse(self.do_parse("1979-02-29")[0])

    def test_date_parser_must_not_accept_illegal_day_germandate(self):
        self.assertFalse(self.do_parse("32.07.1979")[0])
        self.assertFalse(self.do_parse("29.02.1979")[0])

    def test_date_parser_must_not_accept_illegal_day_american_date(self):
        self.assertFalse(self.do_parse("07/32/1979")[0])
        self.assertFalse(self.do_parse("02/29/1979")[0])

    def test_date_parser_must_not_accept_illegal_month_isoformat(self):
        self.assertFalse(self.do_parse("1979-13-18")[0])

    def test_date_parser_must_not_accept_illegal_month_germandate(self):
        self.assertFalse(self.do_parse("18.13.1979")[0])

    def test_date_parser_must_not_accept_illegal_month_american_date(self):
        self.assertFalse(self.do_parse("13/18/1979")[0])

    def test_date_parser_must_accept_leap_year_isoformat(self):
        self.assertEqual(date(2000, 2, 29), self.do_parse("2000-02-29")[0])

    def test_date_parser_must_accept_leap_year_germandate(self):
        self.assertEqual(date(2000, 2, 29), self.do_parse("29.02.2000")[0])

    def test_date_parser_must_accept_leap_year_american_date(self):
        self.assertEqual(date(2000, 2, 29), self.do_parse("02/29/2000")[0])

    def test_date_parser_must_not_accept_non_leap_year_isoformat(self):
        self.assertFalse(self.do_parse("1900-02-29")[0])

    def test_date_parser_must_not_accept_non_leap_year_germandate(self):
        self.assertFalse(self.do_parse("29.02.1900")[0])

    def test_date_parser_must_not_accept_non_leap_year_american_date(self):
        self.assertFalse(self.do_parse("02/29/1900")[0])

    def test_date_parser_must_not_accept_stuff_before_iso_date(self):
        self.assertFalse(self.do_parse("The date is 2020-02-29")[0])

    def test_date_parser_must_not_accept_stuff_before_germandate(self):
        self.assertFalse(
            self.do_parse("Das heutige Datum ist: 29.02.2020")[0],
        )

    def test_date_parser_must_not_accept_stuff_before_american_date(self):
        self.assertFalse(
            self.do_parse("The date is 02/29/2020")[0]
        )

    def test_date_parser_must_accept_stuff_after_iso_date(self):
        self.assertEqual(
            date(2020, 2, 29), self.do_parse("2020-02-29 is the date")[0]
        )

    def test_date_parser_must_accept_stuff_after_germandate(self):
        self.assertEqual(
            date(2020, 2, 29),
            self.do_parse(
                "29.02.2020, der Tag den es eigentlich nicht geben durfte"
            )[0],
        )

    def test_date_parser_must_accept_stuff_after_american_date(self):
        self.assertEqual(
            date(2020, 2, 29), self.do_parse("02/29/2020 is the date")[0]
        )

    def test_date_parser_must_not_match_additional_numbers_before_iso_date(self):
        self.assertFalse(self.do_parse("02020-02-29")[0])

    def test_date_parser_must_not_match_additional_numbers_before_germandate(self):
        self.assertFalse(self.do_parse("02020-02-29")[0])

    def test_date_parser_must_not_match_additional_numbers_before_american_date(self):
        self.assertFalse(self.do_parse("02020-02-29")[0])

    def test_date_parser_must_not_accept_additional_numbers_after_iso_date(self):
        self.assertFalse(self.do_parse("2020-02-290")[0])

    def test_date_parser_must_not_accept_additional_numbers_after_germandate(self):
        self.assertFalse(
            self.do_parse("29.02.200", now_provider=lambda: datetime(2025, 1, 1, 0, 0))[0],
        )

    def test_date_parser_must_not_accept_additional_numbers_after_american_date(self):
        self.assertFalse(self.do_parse("02/29/200")[0])

    def test_date_parser_must_accept_iso_datetime(self):
        self.assertEqual(
            datetime(2020, 2, 29, 12, 34, 56),
            self.do_parse("2020-02-29T12:34:56")[0],
        )

    def test_date_parser_must_accept_iso_datetime_with_microseconds(self):
        self.assertEqual(
            datetime(2020, 2, 29, 12, 34, 56, 789000),
            self.do_parse("2020-02-29T12:34:56.789")[0],
        )

    def test_date_parser_must_ignore_trailing_numbers_in_iso_datetime(self):
        self.assertEquals(
            datetime(2020, 2, 29, 12, 34, 56),
            self.do_parse("2020-02-29T12:34:56.789123")[0],
        )

    def test_date_parser_must_not_match_leading_numbers_in_datetime(self):
        self.assertFalse(self.do_parse("1232020-02-29T12:34:56.789123")[0])

    def test_date_parser_must_accept_iso_datetime_even_without_seconds(self):
        self.assertEqual(
            datetime(2020, 2, 29, 12, 34),
            self.do_parse("2020-02-29T12:34")[0],
        )

    def test_date_parser_must_accept_time_with_hour_and_minutes(self):
        self.assertEqual(
            time(12, 34),
            self.do_parse("12:34")[0],
        )

    def test_date_parser_must_accept_time_with_hour_parsed_leniently(self):
        self.assertEqual(
            time(8, 34),
            self.do_parse("8:34")[0]
        )

    def test_date_parser_must_not_leniently_parse_minutes(self):
        self.assertFalse(self.do_parse("12:4")[0])

    def test_date_parser_must_accept_time_with_hour_and_seconds(self):
        self.assertEqual(
            time(12, 34, 56),
            self.do_parse("12:34:56")[0],
        )

    def test_date_parser_must_not_leniently_parse_seconds(self):
        result, tokens = self.do_parse("12:34:5")
        self.assertEqual(time(12, 34), result)
        self.assertEqual(1, len(tokens))
        self.assertEqual("Text", tokens[0].token_type_name)

    def test_ensure_datetime_must_return_datetime_on_datetime_given(self):
        self.assertEqual(
            datetime(2020, 2, 29, 12, 34),
            ensure_datetime(datetime(2020, 2, 29, 12, 34))
        )

    def test_ensure_datetime_must_return_datetime_with_min_time_on_date_given(self):
        self.assertEqual(
            datetime(2020, 2, 29, 0, 0),
            ensure_datetime(date(2020, 2, 29))
        )

    def test_ensure_datetime_must_return_datetime_with_date_taken_from_nowprovider_on_time_given(self):
        self.assertEqual(
            datetime(2020, 2, 29, 13, 47),
            ensure_datetime(time(13,47), now_provider=lambda: datetime(2020, 2, 29, 1, 34))
        )





from datetime import datetime, date, time, timedelta
from typing import Sequence

from ratisbona_utils.datetime._date_tokenizer import token_definitions
from ratisbona_utils.datetime._datetime_utils import to_last_weekday, to_next_weekday
from ratisbona_utils.functional import first, Provider
from ratisbona_utils.monads import Just, Maybe, Error, Nothing
from ratisbona_utils.monads.monads2 import with_errorhandling
from ratisbona_utils.parsing import get_tokenizer, Token

DEBUG = True


if DEBUG:

    def dprint(*args, **kwargs):
        print(*args, **kwargs)

else:

    def dprint(*args, **kwargs):
        pass


lenient_tokenizer = get_tokenizer(token_definitions)


def lenient_parse_year(
    year_str: str, two_digit_year_century_provider: Provider[datetime] = datetime.now
) -> Maybe[int | ValueError]:
    """
    Leniently parse a year string that may have 2 or 4 digits.
    Args:
        year_str: the year string to parse
        two_digit_year_century_provider: Provider of datetime. This datetime is used to determine the century for 2-digit years.

    Returns:
        Maybe[int | ValueError]: Just(year) if parsing was successful, Error(ValueError) otherwise.

    """
    if len(year_str) == 2:
        now = two_digit_year_century_provider()
        current_year = now.year
        current_century = current_year // 100
        current_centuries_year = current_year % 100
        year_int = int(year_str)
        if year_int > current_centuries_year:
            current_century -= 1
        current_century_str = str(current_century)
        current_century_prefix = current_century_str[:2]
        year_str = current_century_prefix + year_str
    if len(year_str) != 4:
        return Error(ValueError(f"Year must have 2 or 4 digits, but got '{year_str}'"))
    try:
        year_int = int(year_str)
    except ValueError:
        return Error(ValueError(f"Year must be a number, but got '{year_str}'"))
    return Just(year_int)


def try_parse_datetime(token: Token) -> Maybe[datetime | ValueError]:
    """
    Try to parse a datetime from a token like the lenient tokenizer produces, that means it is of type 'Datetime'
    and it has to contain the fields:
        - year
        - month
        - day
        - hour
        - minute
        - second (optional, defaults to 0)
        - millisecond (optional, defaults to 0)

    Note that returning Errors might indicate simply that the token is of another type and another parsing
    function should be tried.

    Args:
        token: A lenient tokenizer like token of type 'Datetime'.

    Returns:
        Maybe[datetime | ValueError]: Just(datetime) if parsing was successful, Error(ValueError) otherwise.

    Returns ValueError if:
        - The token is not of type 'Datetime'.
        - Any of the required fields are missing or cannot be converted to int.

    """
    if token.token_type_name != "Datetime":
        return Error(
            ValueError(
                f"Token is not a datetime token: {token}. Expected Type 'Datetime' but got '{token.token_type_name}'"
            )
        )
    maybe_match_groups = Just(
        token.token_match_groups
    )  # Will throw for a Datetime Token without match groups!
    maybe_year = maybe_match_groups["year"].bind(int)
    maybe_month = maybe_match_groups["month"].bind(int)
    maybe_day = maybe_match_groups["day"].bind(int)
    maybe_hour = maybe_match_groups["hour"].bind(int)
    maybe_minute = maybe_match_groups["minute"].bind(int)
    maybe_second = maybe_match_groups["second"].bind(int).default_or_error(0)
    maybe_microseconds = (
        maybe_match_groups["millisecond"]
        .bind(int)
        .default_or_error(0)
        .bind(lambda millisecond: millisecond * 1000)
    )
    return with_errorhandling(
        datetime,
        maybe_year,
        maybe_month,
        maybe_day,
        maybe_hour,
        maybe_minute,
        maybe_second,
        maybe_microseconds,
    )


def try_parse_date(
    token: Token, implicit_year_provider: Provider[datetime] = datetime.now
) -> Maybe[date | ValueError]:
    """
    Try to parse a date from a token like the lenient tokenizer produces, that means it is of type 'Date'
    and it has to contain the fields:
        - year (optional, maybe 2 maybe 4 digits, defaults year from now_provider)
        - month
        - day

    Args:
        token: The lenient tokenizer like token of type 'Date'.
        implicit_year_provider: The provider of the current datetime, used to get the current year if the year is missing.

    Returns:

    """
    if token.token_type_name != "Date":
        return Error(
            ValueError(
                f"Token is not a date token: {token}. Expected Type 'Date' but got '{token.token_type_name}'"
            )
        )
    maybe_match_groups = Just(
        token.token_match_groups
    )  # Will throw for a Date Token without match groups!
    maybe_year = (
        maybe_match_groups["year"]
        .bind(lenient_parse_year, implicit_year_provider)
        .default_or_error(implicit_year_provider().year)
    )
    maybe_month = maybe_match_groups["month"].bind(int)
    maybe_day = maybe_match_groups["day"].bind(int)
    return with_errorhandling(date, maybe_year, maybe_month, maybe_day)


def try_parse_time(token: Token) -> Maybe[time | ValueError]:
    """
    Try to parse a time from a token like the lenient tokenizer produces, that means it is of type 'Time'
    and it has to contain the fields:
        - hour
        - minute
        - second (optional, defaults to 0)
        - millisecond (optional, defaults to 0)

    Args:
        token: A lenient tokenizer like token of type 'Time'.

    Returns:
        Maybe[time | ValueError]: Just(time) if parsing was successful, Error(ValueError) otherwise.

    """
    if token.token_type_name != "Time":
        return Error(
            ValueError(
                f"Token is not a time token: {token}. Expected Type 'Time' but got '{token.token_type_name}'"
            )
        )
    maybe_match_groups = Just(
        token.token_match_groups
    )  # Will throw for a Time Token without match groups!
    maybe_hour = maybe_match_groups["hour"].bind(int)
    maybe_minute = maybe_match_groups["minute"].bind(int)
    maybe_second = maybe_match_groups["second"].bind(int).default_or_error(0)
    maybe_microseconds = (
        maybe_match_groups["millisecond"]
        .bind(int)
        .default_or_error(0)
        .bind(lambda millisecond: millisecond * 1000)
    )
    return with_errorhandling(
        time, maybe_hour, maybe_minute, maybe_second, maybe_microseconds
    )


def try_parse_relative_date(
    token: Token, relative_to_now_provider: Provider[datetime] = datetime.now
) -> Maybe[date | ValueError]:
    """
    Parses a relative date token and returns the corresponding date based on the given
    reference date provider. Handles parsing of relative terms such as "today,"
    "tomorrow," "yesterday," or specific weekdays.

    Parameters:
    token (Token): The token to parse. It must be of type 'RelativeDate', and its
                   match groups should include relevant keys for relative date processing.
    now_provider (Provider[datetime]): A callable that provides the current datetime
                                       used as point of reference for date calculations.
                                       Default is datetime.now.

    Returns:
    Maybe[date | ValueError]: Returns a Maybe-wrapped date object if parsing is
                              successful, or a Maybe-wrapped ValueError if the token
                              does not represent a valid relative date.

    Raises:
    ValueError: Raised when the input token is not of type 'RelativeDate' or contains
                unknown relative date formats.
    """
    if token.token_type_name != "RelativeDate":
        return Error(
            ValueError(
                f"Token is not a relative date token: {token}. Expected Type 'RelativeDate' but got '{token.token_type_name}'"
            )
        )
    maybe_match_groups = Just(
        token.token_match_groups
    )  # Will throw for a RelativeDate Token without match groups!
    rel = maybe_match_groups["rel"].bind(str.lower).unwrap_value()
    now = relative_to_now_provider()
    if rel in ["heute", "today"]:
        return Just(now.date())
    if rel in ["morgen", "tomorrow"]:
        return Just(now.date() + timedelta(days=1))
    if rel in ["übermorgen", "day after tomorrow"]:
        return Just(now.date() + timedelta(days=2))
    if rel in ["gestern", "yesterday"]:
        return Just(now.date() - timedelta(days=1))
    if rel in ["vorgestern", "day before yesterday"]:
        return Just(now.date() - timedelta(days=2))

    last = any([keyword in rel for keyword in ["letzter", "last"]])

    # noformat: on
    to_weekday_table = [
        (["montag", "monday", "mon", "mo"], 0),  # Monday
        (["dienstag", "tuesday", "tue", "di"], 1),  # Tuesday
        (["mittwoch", "wednesday", "wed", "mi"], 2),  # Wednesday
        (["donnerstag", "thursday", "thu", "do"], 3),  # Thursday
        (["freitag", "friday", "fri", "fr"], 4),  # Friday
        (["samstag", "saturday", "sat", "sa"], 5),  # Saturday
        (["sonntag", "sunday", "sun", "so"], 6),  # Sunday
    ]
    # noformat: off

    last_word_of_rel = rel.split()[-1].strip()
    for weekdays, weekday_index in to_weekday_table:
        if last_word_of_rel in weekdays:
            weekday_timespan_function = to_last_weekday if last else to_next_weekday
            nowdate = relative_to_now_provider()
            the_date = nowdate + weekday_timespan_function(nowdate, weekday_index)
            return Just(the_date)

    return Error(ValueError(f"Unknown relative date: '{rel}'"))


def _maybe_eat_separator_tokens(tokens: Sequence[Token]) -> tuple[Sequence[Token], list[Token]]:
    eaten_tokens = []
    while len(tokens) > 0 and tokens[0].token_type_name == "Pure Separators":
        dprint(f"Eating separator token: {tokens[0]}")
        eaten_tokens.append(tokens[0])
        tokens = tokens[1:]
    return tokens, eaten_tokens

def _format_tokenlist(tokens: Sequence[Token], separator="\n\t") -> str:
    result = ""
    for idx, token in enumerate(tokens):
        if idx > 0:
            result += separator
        result += str(token)
    return result

def leniently_parse(
    tokens: Sequence[Token],
    *,
    base_date_for_implicit_dates: Provider[datetime] = datetime.now,
    base_date_for_relative_dates: Provider[datetime] = datetime.now
) -> tuple[Maybe[date | time | datetime], Sequence[Token]]:
    """
        Facade-Routine that tries to parse a date, time or datetime from a sequence of tokens.

    """
    if len(tokens) == 0:
        return Nothing, []

    tokens = list(tokens)
    eaten_tokens = []
    next_token = tokens.pop(0)

    if next_token.token_type_name == "Datetime":
        maybe_datetime = try_parse_datetime(next_token)
        if maybe_datetime:
            dprint(f"If we found a datetime all is clear: {maybe_datetime}")
            return maybe_datetime, tokens

    maybe_date = Nothing
    if next_token.token_type_name == "Date":
        maybe_date = try_parse_date(next_token, implicit_year_provider=base_date_for_implicit_dates)

        if maybe_date:
            dprint(f"The token is a date {maybe_date}...")
            tokens, eaten_tokens = _maybe_eat_separator_tokens(tokens)
            if not tokens:
                dprint(f"...and there are no more relevant tokens, so we are done.")
                return maybe_date, eaten_tokens
            next_token = tokens.pop(0)
            dprint(
                f"...but if there are more tokens, we need to check the next token: {next_token}"
            )

    # If we did not find a date yet, we check if the next token is a relative date.
    if not maybe_date and next_token.token_type_name == "RelativeDate":
        maybe_date = try_parse_relative_date(next_token, relative_to_now_provider=base_date_for_relative_dates)
        if maybe_date:
            dprint(f"We found a relative date: {maybe_date}...")
            tokens, eaten_tokens = _maybe_eat_separator_tokens(tokens)
            if not tokens:
                dprint(
                    f"...and we consumed all the relevant tokens, now we are done."
                )
                return maybe_date, eaten_tokens
            next_token = tokens.pop(0)
            dprint(
                f"...but there are more tokens, "
                f"we need to check the next token: {next_token}"
            )

    maybe_time = Nothing
    if next_token.token_type_name == "Time":
        maybe_time = try_parse_time(next_token)
        if maybe_time:
            dprint(
                f"We found a time {maybe_time}, so we consumed the token without pulling another one."
            )
            next_token = Nothing
            eaten_tokens = []

    if next_token or eaten_tokens:
        dprint(
            "We either did not match anything or we matched and then pulled another token,\n"
            "in which case we need to put it back into the unparsed tokens."
        )

        if next_token:
            tokens.insert(0, next_token)
        for eaten_token in reversed(eaten_tokens):
            tokens.insert(0, eaten_token)
        next_token = Nothing
        eaten_tokens = []
        dprint(f"...so we put the token(s) back into the unparsed tokens:\n{_format_tokenlist(tokens)}")

    if maybe_date and maybe_time:
        the_datetime = maybe_date.bind(datetime.combine, maybe_time)
        dprint(
            f"We found a date and a time, we combine them into a datetime: {the_datetime}. Leftover tokens:\n\t{_format_tokenlist(tokens)}"
        )
        assert(len(eaten_tokens) == 0) # there can not be any eaten tokens in that case, because parse time does not skip over tokens.
        return the_datetime, tokens

    if maybe_date and not maybe_time:
        dprint(
            f"We found a date but no time, we return the date: {maybe_date}. Leftover tokens:\n\t{_format_tokenlist(tokens)}"
        )
        for eaten_token in reversed(eaten_tokens):
            tokens.insert(0, eaten_token)
        return maybe_date, tokens

    if not maybe_date and maybe_time:
        dprint(
            f"We found a time but no date, we return the time: {maybe_time}. Leftover tokens:\n\t{_format_tokenlist(tokens)}"
        )
        assert (len(eaten_tokens) == 0)  # there can not be any eaten tokens in that case, because parse time does not skip over tokens.
        return maybe_time, tokens

    print("Did not find any datelike stuff... Well, shit...")
    for eaten_token in reversed(eaten_tokens):
        tokens.insert(0, eaten_token)
    return Nothing, tokens


def ensure_datetime(
    value: datetime | date | time, now_provider: Provider[datetime] = datetime.now
) -> datetime:
    if isinstance(value, datetime):
        return value
    if isinstance(value, date):
        return datetime.combine(value, time.min)
    if isinstance(value, time):
        return datetime.combine(now_provider().date(), value)
    raise ValueError(f"Expected a datetime, date or time object, but got '{value}'")


if __name__ == "__main__":
    while True:
        print("Input:", end=" ")
        the_input = input()
        tokens = list(lenient_tokenizer(the_input))
        for token in tokens:
            print(token)

        the_parseresult = leniently_parse(tokens)
        print(the_parseresult, end="")
        print(f" {type(the_parseresult.unwrap_value())}" if the_parseresult else "")



from typing import List, Callable, Any



def nullsafe_decorator(
    exclude_indices: List[int] = [], exclude_keys: List[str] = []
) -> Callable[[Callable[[...], Any]], Callable[[...], Any]]:
    """
    Decorator to make a function nullsafe.

    Args:
        exclude_indices:
            List of indices to exclude from the nullsafe check. If an argument at that index is None,
            the function will still be called.
        exclude_keys:
            List of keys to exclude from the nullsafe check. If a keyword argument with that
            key is None, the function will still be called.
    """

    def decorator(function: Callable[[...], Any]) -> Callable[[...], Any]:
        def wrapper(*args, **kwargs):
            for idx, arg in enumerate(args):
                if arg is None and idx not in exclude_indices:
                    return None
            for key, value in kwargs.items():
                if value is None and key not in exclude_keys:
                    return None
            return function(*args, **kwargs)

        return wrapper

    return decorator


def nullsafe(
    function: Callable[[...], Any],
    exclude_indices: List[int] = [],
    exclude_keys: List[str] = [],
) -> Callable[[...], Any]:
    """
    Make a function nullsafe. Convenience function for nullsafe_decorator.

    Args:
        function:
            The function to make nullsafe.
        exclude_indices:
            List of indices to exclude from the nullsafe check. If an argument at that index is None,
            the function will still be called.
        exclude_keys:
            List of keys to exclude from the nullsafe check. If a keyword argument with that
            key is None, the function will still be called.
    """
    return nullsafe_decorator(exclude_indices, exclude_keys)(function)


def saveget(dictionary, key, default=None):
    if not dictionary:
        return default
    return dictionary.get(key, default)


def substitute(value, substitute):
    return value if value else substitute


def has_keyvalue(dictionary, key):
    return dictionary.get(key, None) is not None


import unittest


class LatexTestCase(unittest.TestCase):
    def test_make_text_to_latex_paragraphs_must_emit_paragraphs(self):
        from ratisbona_utils.latex.linebreaks import make_text_to_latex_paragraphs

        input_text = "Hello!\nTest!\nThis is a test.\n\nThis is another paragraph.\n\n\nThis is yet another paragraph."
        expected_output = "Hello!\n\nTest!\n\nThis is a test.\n\nThis is another paragraph.\n\nThis is yet another paragraph.\n"
        actual_output = make_text_to_latex_paragraphs(input_text)
        self.assertEqual(expected_output, actual_output)


if __name__ == '__main__':
    unittest.main()



import base64

import magic


def encode_base64_embedded(imageData: bytes) -> str:
    """
    Encode the image data in base64 and return the string to be used in html
    Args:
        imageData: the image data to be encoded
    Returns:
        the base64 encoded image
    """
    themagic = magic.from_buffer(imageData, mime=True)
    if not themagic.startswith('image/'):
        raise ValueError(f"Data is not an image. Magic says: {themagic}")
    return f"data:{themagic};base64,{base64.b64encode(imageData).decode('utf-8')}"


"""
    This module provides a set of classes and functions to work with colors.
    It can handle the following color spaces: RGB, normalized RGB,
    linearized sRGB, XYZ, Lab, and hexadecimal.
    It also provides functions to convert between these color spaces.
    It also provides functions to interpolate between colors.
    It also provides functions to add colors.
    It also provides functions to blend colors.
"""

from . import resources
from .serde_gimp import to_gimp_palette, parse_gimp_palette, to_gimp_gradient
from .palette import Palette, linear_interpolate, linear_interpolate_two_colors
from .simple_color import (
    RGBComponent,
    RGBColor,
    NormalizedRGBComponent,
    NormalizedRGBColor,
    LinearizedsRGBComponent,
    LinearizedsRGBColor,
    XYZComponent,
    XYZColor,
    LabLComponent,
    LababComponent,
    LabColor,
    HexColor,
    hex_to_rgb,
    rgb_to_hex,
    rgb_to_normalized_component,
    normalized_to_rgb_component,
    rgb_to_normalized,
    normalized_to_rgb,
    gamma_function,
    inverse_gamma_function,
    normalized_to_linearized,
    linearized_to_normalized,
    add_colors,
    add_rgb_colors,
    add_normalized_rgb_colors,
    add_linearized_srgb_colors,
    multiply_color,
    multiply_normalized_rgb_color,
    multiply_linearized_srgb_color,
    linear_blend,
    linear_blend_rgb,
    srgb_linearized_to_uxyz,
    uxyz_color_to_linearized_srgb_color,
    D50_XYZ_2deg,
    D65_XYZ_10deg,
    D50_XYZ_10deg,
    D65_XYZ_2deg,
    linearize_uxyz_color_relative_to_whitepoint,
    delinearize_uxyz_color_relative_to_whitepoint,
    uxyz_color_to_lab_color,
    lab_color_to_uxyz_color,
    rgb_to_lab,
    lab_to_rgb,
)



def quote_for_bash_double_quotes(
    string: str,
    escape_double_quotes: bool = True,
    escape_backslashes: bool = True,
    escape_dollarsigns: bool = True,
    escape_backticks: bool = True,
    escape_exclamation_marks: bool = True,
) -> str:
    """
    Quote a string for use in bash with double quotes.

    Args:
        string (str): The string to quote.
        escape_double_quotes (bool): Whether to escape double quotes in the string.
        escape_backslashes (bool): Whether to escape backslashes in the string.
        escape_dollarsigns (bool): Whether to escape dollar signs in the string.
        escape_backticks (bool): Whether to escape backticks in the string.
        escape_exclamation_marks (bool): Whether to escape exclamation marks in the string.

    Default:
        All flags true!

    Returns:
        str: The quoted string, einclosed in double quotes.
    """
    if escape_double_quotes:
        string = string.replace('"', '\\"')
    if escape_backslashes:
        string = string.replace('\\', '\\\\')
    if escape_dollarsigns:
        string = string.replace('$', '\\$')
    if escape_backticks:
        string = string.replace('`', '\\`')
    if escape_exclamation_marks:
        string = string.replace('!', '\\!')

    return f'"{string}"'


"""
    This module contains utility functions for the English language.
"""

def encode_english_short_ordinal(ith: int) -> str:
    """
        Encodes the given ordinal number as a short ordinal string.

        Examples:
            1 -> 1st
            2 -> 2nd
            ...
            10 -> 10th
            11 -> 11th
            ...
            20 -> 20th
            21 -> 21st
            ...
        Args:
            ith: The ordinal number to encode.

        Return:
            The encoded ordinal number.
         """
    if 4 <= ith % 100 <= 20 or 4 <= ith % 10 <= 9 or ith % 10 == 0:
        return f'{ith}th'
    return f'{ith}'+['st', 'nd', 'rd'][ith % 10 - 1]


from io import StringIO
from pathlib import Path

from ratisbona_utils.colors.simple_color import RGBColor, NormalizedRGBColor
from ratisbona_utils.colors.palette import Palette


def to_gimp_palette(palette: Palette[RGBColor]) -> str:
    """
    Convert a Palette object to a GIMP palette file format.

    Args:
        palette (Palette): The palette object to convert

    Returns:
        str: The GIMP palette file content
    """
    description = palette.description.replace("\n", " ")
    lines = [
        "GIMP Palette",
        f"Name: {palette.name}",
        "Columns: 16",
        f"# Description: {description}",
        f"# Author: {palette.author}",
        f"# Created: {palette.creation_date}",
    ]
    for color, name in zip(palette.colors, palette.color_names):
        lines.append(f"{color[0]} {color[1]} {color[2]} {name}")
    return "\n".join(lines)


def parse_gimp_palette(palette_as_str: str) -> Palette:
    """
    Parse a GIMP palette file and return the Palette object.

    Args:
        palette_as_str (str): The GIMP palette file content as a string

    Returns:
        Palette: The parsed palette object
    """
    lines = palette_as_str.splitlines()
    lines = map(str.strip, lines)
    lines = list(filter(lambda s: s and not s.startswith("#"), lines))
    if not lines[0] == "GIMP Palette":
        raise ValueError("Not a GIMP palette file")
    lines = lines[1:]
    name = "?"
    while True:
        current_line = lines[0]

        if current_line.startswith("Name: "):
            name = current_line[6:]
            lines = lines[1:]
            continue

        if current_line.startswith("Columns: "):
            lines = lines[1:]
            continue

        break

    colors = []
    names = []
    for line in lines:
        if not line:
            continue
        color = line.split()
        colors.append((int(color[0]), int(color[1]), int(color[2])))
        names.append(color[3] if len(color) > 3 else "")
    return Palette(
        name=name,
        description="",
        author="",
        creation_date=None,
        colors=colors,
        color_names=names,
        color_types=["rgb"] * len(colors),
    )


def to_gimp_gradient(title: str, colors: list[NormalizedRGBColor]) -> str:
    """Generate a GIMP .ggr gradient file from a list of normalized RGB tuples."""
    num_segments = len(colors) - 1  # Number of segments

    if num_segments < 1:
        raise ValueError("At least two colors are needed to create a gradient.")

    f = StringIO()
    f.write("GIMP Gradient\n")
    f.write(f"Name: {title}\n")
    f.write(f"{num_segments}\n")  # Number of segments

    for i in range(num_segments):
        r1, g1, b1 = colors[i]
        r2, g2, b2 = colors[i + 1]
        start = i / num_segments
        end = (i + 1) / num_segments
        middle = (start + end) / 2  # Linear interpolation for midpoint

        # Format: start middle end  R1 G1 B1 A1  R2 G2 B2 A2  blend_type color_type
        f.write(
            f"{start:.6f} {middle:.6f} {end:.6f}   {r1:.6f} {g1:.6f} {b1:.6f} 1.0   {r2:.6f} {g2:.6f} {b2:.6f} 1.0   0 0\n")
    return f.getvalue()





from typing import Optional

class State:
    def transition(self, *args, **kwargs) -> Optional['State']:
        """Determine the next state. Return None to stay in the current state."""
        return None

    def state_entered(self):
        """Called after entering this state."""
        pass

    def state_left(self):
        """Called before leaving this state."""
        pass


class StateMachine:
    def __init__(self, initial_state: State):
        self.current_state = initial_state
        self.current_state.state_entered()

    def __getattr__(self, name: str):
        state_attr = getattr(self.current_state, name, None)

        if not callable(state_attr):
            raise AttributeError(f"'{type(self).__name__}' and current state '{type(self.current_state).__name__}' have no attribute '{name}'")

        def wrapper(*args, **kwargs):
            result = state_attr(*args, **kwargs)

            # Determine next state
            next_state = self.current_state.transition(*args, **kwargs)
            if next_state and next_state != self.current_state:
                self.current_state.state_left()
                self.current_state = next_state
                self.current_state.state_entered()

            return result

        return wrapper



def lookahead(pattern: str) -> str:
    """
    This function takes a regex pattern and returns a modified pattern that uses lookahead assertions.

    Lookahead matches if the next characters match the given pattern, without consuming any these characters.

    For example `Isaac (?=Asimov)` matches `Isaac ` only if it is followed by `Asimov`, but does not consume `Asimov`.

    Args:
        pattern (str): The regex pattern to modify.

    Returns:
        str: The modified regex pattern with lookahead assertions.
    """

    # Add lookahead assertions at the beginning and end of the pattern
    return '(?=' + pattern + ')'

def negative_lookahead(pattern: str) -> str:
    """
    This function takes a regex pattern and returns a modified pattern that uses negative lookahead assertions.

    Negative lookahead matches if the next characters do not match the given pattern, without consuming any these characters.

    For example `Isaac (?!Asimov)` matches `Isaac ` only if it is NOT followed by `Asimov`, but does not consume `Asimov`.

    Args:
        pattern (str): The regex pattern to modify.

    Returns:
        str: The modified regex pattern with negative lookahead assertions.
    """

    # Add negative lookahead assertions at the beginning and end of the pattern
    return '(?!' + pattern + ')'

def lookbehind(pattern: str) -> str:
    """
    This function takes a regex pattern and returns a modified pattern that uses lookbehind assertions.

    Lookbehind matches if the previous characters match the given pattern, without consuming any these characters.

    For example `(?<=Isaac) Asimov` matches `Asimov` only if it is preceded by `Isaac`, but does not consume `Isaac`.

    The contained pattern must only match strings of a fixed length.
    For example `abc` and `a|b` are allowed, but things like `a*` or `a+` or `a{3,4}`are not.

    Args:
        pattern (str): The regex pattern to modify.

    Returns:
        str: The modified regex pattern with lookbehind assertions.
    """

    # Add lookbehind assertions at the beginning and end of the pattern
    return '(?<' + pattern + ')'

def negative_lookbehind(pattern: str) -> str:
    """
    This function takes a regex pattern and returns a modified pattern that uses negative lookbehind assertions.

    Negative lookbehind matches if the previous characters do not match the given pattern, without consuming any these characters.
    For example `(?!Isaac) Asimov` matches `Asimov` only if it is NOT preceded by `Isaac`, but does not consume `Isaac`.

    The contained pattern must only match strings of a fixed length.

    Args:
        pattern (str): The regex pattern to modify.

    Returns:
        str: The modified regex pattern with negative lookbehind assertions.
    """
    return '(?<!' + pattern + ')'

def with_word_boundaries(pattern: str) -> str:
    """
        Surrounds the given pattern with word boundaries.

        Args:
            pattern (str): The regex pattern to modify.

        Returns:
            str: The modified regex pattern with word boundaries.
    """
    return r'\b' + pattern + r'\b'

def named_matchgroup(name: str, pattern: str) -> str:
    """
        Creates a named match group for the given pattern.

        Args:
            name (str): The name of the match group.
            pattern (str): The regex pattern to modify.

        Returns:
            str: The modified regex pattern with a named match group.
    """
    return f'(?P<{name}>{pattern})'

def non_capturing_group(pattern: str) -> str:
    """
        Creates a non-capturing group for the given pattern.

        Args:
            pattern (str): The regex pattern to modify.

        Returns:
            str: The modified regex pattern with a non-capturing group.
    """
    return '(?:' + pattern + ')'

def any_of(*patterns: str) -> str:
    """
        Creates a regex pattern that matches any of the given patterns.

        Args:
            *patterns (str): The regex patterns to combine.

        Returns:
            str: The combined regex pattern.
    """
    return "[" + "".join(patterns) + "]"

def optional(pattern: str) -> str:
    """
        Creates a regex pattern that matches the given pattern zero or one time.

        Args:
            pattern (str): The regex pattern to modify.

        Returns:
            str: The modified regex pattern with optional matching.
    """
    return f"({pattern})?"


import unittest

from ratisbona_utils.arrays import D2Array



def generate_test_array():
    return D2Array.by_generating_func(lambda r, c: (r, c), 9, 5)


class D2ArrayTestCase(unittest.TestCase):

    def test_iterator_must_traverse_array_in_row_major_order(self):
        the_test_array = generate_test_array()
        expected_cells = []
        for row in range(9):
            for col in range(5):
                expected_cells.append((row, col))
        actual_cells = list(the_test_array)
        self.assertEqual(expected_cells, actual_cells)

    def test_slicing_must_produce_correct_subarray(self):
        the_test_array = generate_test_array()
        subarray = the_test_array[2:7:2, 1:5:2]
        expected_cells = []
        for row in range(2, 7, 2):
            for col in range(1, 5, 2):
                expected_cells.append((row, col))
        actual_cells = list(subarray)
        self.assertEqual(expected_cells, actual_cells)

    def test_slicing_must_allow_element_access(self):
        the_test_array = generate_test_array()
        for row in range(9):
            for col in range(5):
                expected = (row, col)
                actual = next(iter(the_test_array[row, col]))
                print(f"Checking [{row}, {col}]: {actual}")
                self.assertEqual(expected, actual)

    def test_slicing_must_raise_value_error_on_invalid_slices(self):
        the_test_array = generate_test_array()
        with self.assertRaises(ValueError) as cm:
            _ = the_test_array[0:0, 0:5]
        print(cm.exception)

    def test_slicing_must_raise_value_error_on_invalid_slice_types(self):
        the_test_array = generate_test_array()
        with self.assertRaises(ValueError) as cm:
            _ = the_test_array["invalid", 0:5]
        print(cm.exception)
        with self.assertRaises(ValueError) as cm:
            _ = the_test_array[0:5, "invalid"]
        print(cm.exception)

    def test_slicing_must_raise_value_error_on_invalid_tuple_length(self):
        the_test_array = generate_test_array()
        with self.assertRaises(ValueError) as cm:
            _ = the_test_array[0:5, 0:5, 0:5]
        print(cm.exception)

    def test_slicing_single_row_and_column(self):
        the_test_array = generate_test_array()
        subarray = the_test_array[4]
        expected_cells = []
        for col in range(5):
            expected_cells.append((4, col))
        actual_cells = list(subarray)
        self.assertEqual(expected_cells, actual_cells)

    def test_slicing_must_clone_array(self):
        the_test_array = generate_test_array()
        subarray = the_test_array[:, :]
        expected_cells = list(the_test_array)
        actual_cells = list(subarray)
        self.assertEqual(expected_cells, actual_cells)


    def test_diagonal_iterator(self):
        the_test_array = generate_test_array()
        diagonal_elements = list(the_test_array.diagonal_iterator())
        expected_elements = [(i, i) for i in range(min(the_test_array.num_rows, the_test_array.num_cols))]
        self.assertEqual(expected_elements, diagonal_elements)

    def test_diagonal_numbers(self):
        the_test_array = generate_test_array() # 9 rows, 5 cols

        min_diagonal_index = -(the_test_array.num_rows - 1)
        max_diagonal_index = the_test_array.num_cols - 1

        self.assertEqual(-8, min_diagonal_index)
        self.assertEqual(4, max_diagonal_index)

    def test_lower_diagonals_without_wrapping(self):
        the_test_array = generate_test_array()
        for diag_index in range(-8, 0):
            expected_elements = []
            row_start = max(0, -diag_index)
            col_start = 0
            while row_start < the_test_array.num_rows and col_start < the_test_array.num_cols:
                expected_elements.append((row_start, col_start))
                row_start += 1
                col_start += 1
            actual_elements = list(the_test_array.diagonal_iterator(diag_index, wrap_around=False))
            self.assertEqual(expected_elements, actual_elements)

    def test_upper_diagonals_without_wrapping(self):
        the_test_array = generate_test_array()
        for diag_index in range(1, 5):
            expected_elements = []
            row_start = 0
            col_start = diag_index
            while row_start < the_test_array.num_rows and col_start < the_test_array.num_cols:
                expected_elements.append((row_start, col_start))
                row_start += 1
                col_start += 1
            actual_elements = list(the_test_array.diagonal_iterator(diag_index, wrap_around=False))
            self.assertEqual(expected_elements, actual_elements)

    def test_lower_diagonals_with_wrapping(self):
        the_test_array = generate_test_array()
        for diag_index in range(-8, 0):
            expected_elements = []
            row_start = max(0, -diag_index)
            col_start = 0
            while len(expected_elements) < the_test_array.num_rows:
                expected_elements.append((row_start % the_test_array.num_rows, col_start % the_test_array.num_cols))
                row_start += 1
                col_start += 1
            actual_elements = list(the_test_array.diagonal_iterator(diag_index, wrap_around=True))
            print(f"Diagonal {diag_index} with wrapping: expected \n{expected_elements}, got \n{actual_elements}")
            self.assertEqual(expected_elements, actual_elements)







if __name__ == '__main__':
    unittest.main()



import os
import shutil
import sys
from importlib import resources
from importlib.abc import Traversable
from pathlib import Path

UTF8 = {"encoding": "utf-8"}


def errprint(*args, **kwargs):
    """
    Prints to stderr.

    Args:
        *args: Arguments to print.
        **kwargs: Keyword arguments to print.

    Returns:
        None

    Side Effects:
        - Prints to stderr.
    """
    print(*args, file=sys.stderr, **kwargs)


def maybe_backup_file(filepath: Path) -> bool:
    if not filepath.exists():
        return False
    print(f"Backuping {filepath}")
    with filepath.with_suffix(".bak").open("wb") as backup, filepath.open("rb") as file:
        backup.write(file.read())
    return True


def get_config_dir(toolname="ratisbona_utils", ensure=True, ensure_ratisbona=True):
    """
    Returns the configuration directory for a tool.

    Args:
        toolname (str): The name of the tool. Default: 'ratisbona_utils'.
        ensure (bool): If True, create the configuration directory if it does not exist.

    Returns:
        Path of the configuration directory.

    Side Effects:
        - Maybe creates the configuration directory, if it does not exist.
    """
    if ensure_ratisbona and not toolname.lower().startswith("ratisbona_"):
        toolname = f"ratisbona_{toolname}"

    config_dir = Path(os.path.expanduser(f"~/.local/share/{toolname}"))
    if ensure:
        config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir

def get_resource(package: str, resource_name: str) -> Traversable:
    """
    Returns a stream to read a resource file from a package.

    Args:
        package (str): The package containing the resource (e.g., 'my_package.resources').
        resource_name (str): The name of the resource file (e.g., 'example.txt').

    Returns:
        A file-like object to read the resource content.

    Side Effects:
        - None

    """
    return resources.files(package).joinpath(resource_name)

def copy_resource_file(
    package: str, resource_name: str, output_filename: Path, chunk_size: int = 4096
):
    """
    Reads a resource file from a package and copies it chunkwise to an output file.

    Args:

        package (str): The package containing the resource (e.g., 'my_package.resources').
        resource_name (str): The name of the resource file (e.g., 'example.txt').
        output_file (Path): The destination file to which the content will be written.
        chunk_size: The size of chunks to copy at a time (default: 4096 bytes).

    Returns:
        None

    Side Effects:
        - Copies the resource file to the output file.

    """
    with (
        resources.files(package).joinpath(resource_name).open("rb") as src,
        open(output_filename, "wb") as dst,
    ):
        shutil.copyfileobj(src, dst, length=chunk_size)



import unittest

from ratisbona_utils.colors.simple_color import f_inv, f, lms_color_to_uxyz_color, uxyz_color_to_lms_color


class SimpleColorTests(unittest.TestCase):

    def test_uxyz_color_to_lms_must_be_inverse(self):
        random_xyz_tuples = [
            (0.1, 0.2, 0.3),
            (0.4, 0.5, 0.6),
            (0.7, 0.8, 0.9),
            (0.1, 0.2, 0.3),
            (0.4, 0.5, 0.6),
            (0.7, 0.8, 0.9),
            (0.1, 0.2, 0.3),
            (0.4, 0.5, 0.6),
            (0.7, 0.8, 0.9),
            (0.1, 0.2, 0.3),
        ]
        for random_tuple in random_xyz_tuples:
            with self.subTest(f"Testing {random_tuple}"):
                x, y, z = lms_color_to_uxyz_color(random_tuple)
                l, m, s = uxyz_color_to_lms_color((x, y, z))
                self.assertAlmostEqual(l, random_tuple[0])

    def test_finv_must_be_inv_of_f(self):
        t=0.0
        while t<100.0:
            fval = f(t)
            finvval = f_inv(fval)
            print(t, fval, finvval)
            self.assertAlmostEqual( t, finvval)
            t += 0.01


if __name__ == '__main__':
    unittest.main()
